/* ***********************************************  */
/*   Projet  : Recherche de solution de l'Ane rouge */
/*   Date    : avril 2009                           */
/*   Auteur  : Climoo                               */
/*   Fichier : kl_lectureFichier.c                  */
/*   Description : Module charge la lecture du      */
/*     fichier des donnees du jeu.                  */
/*                                                  */
/* **************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "allocations.h"
#include "kl_definitions.h"
#include "kl_lectureFichier.h"

/*Le nombre de caractere que contient un fichier normalement ecrit*/
#define TAILLE_CORRECTE_FIC (KL_SURFACE+KL_HAUTEUR)

/*Taille du tampon utilise pour la lecture du fichier*/
#define TAILLE_TAMPON_FIC (TAILLE_CORRECTE_FIC+1)

/*Les erreurs pouvant survenir lors de la lecture d'un fichier*/
/*Chaque erreur est composee d'un numero et d'une chaine de caracteres*/

/*Erreur lors de l'ouverture du fichier*/
#define ERR_OUVERTURE 1
#define CH_ERR_OUVERTURE "Impossible d'ouvrir le fichier en lecture\n"

/*Erreur si le nombre de caracteres du fichier n'est pas le bon*/
#define ERR_TAILLE 2
#define CH_ERR_TAILLE "Taille du fichier incorrecte\n"

/*Erreur si le nombre d'occurences (correspondant a la taille) n'est pas le bon*/
#define ERR_NB_SYMBOLES 3
#define CH_ERR_NB_SYMBOLES "Nombre d'occurence incorrecte pour cette piece\n"

/*Erreur si un caractere est incoherent par rapport a ceux permis dans le fichier*/
#define ERR_CARAC_FICHIER 4
#define CH_ERR_CARAC_FICHIER "Caractere non autorise\n"

/*Erreur si une piece n'a pas la bonne forme*/
#define ERR_FORME 5
#define CH_ERR_FORME "Mauvaise forme de piece\n"


/*Provoque un affichage sur la sortie standard des erreurs selon le type d'erreur detectee.
  La chaine de caractere en parametre est un complement sur l'erreur trouvee, qui sera affiche egalement.
  Ce parametre peut etre remplace par NULL si on le souhaite.
  Exemple: Probleme d'ouverture, cplt vaut le nom du fichier*/
static void afficherErreur(int numeroErreur,const char *cplt);

/*Verifie le contenu d'un fichier.
 Affiche des erreurs si:
 -Les caracteres donnes ne correspondent a rien
 -Une piece quelconque n'a pas la bonne taille
 -Une piece quelconque a un format incorrect */
static char verifierContenuSituation(KL_Situation *kl);

/*Initialise une situation Klotski a partir de la chaine de caractere luee dans le fichier.*/
static KL_Situation *initialiserKlotski(char *lecture);

static void afficherErreur(int numeroErreur,const char *cplt){
  fprintf(stderr,"Klotski Erreur : ");
  if (cplt != NULL)
    fprintf(stderr,"%s : ",cplt);
  switch(numeroErreur){
  case ERR_OUVERTURE:
    fprintf(stderr,CH_ERR_OUVERTURE);
    break;
  case ERR_TAILLE:
    fprintf(stderr,CH_ERR_TAILLE);
    break;
  case ERR_NB_SYMBOLES:
    fprintf(stderr,CH_ERR_NB_SYMBOLES);
    break;
  case ERR_CARAC_FICHIER:
    fprintf(stderr,CH_ERR_CARAC_FICHIER);
    break;
  case ERR_FORME:
    fprintf(stderr,CH_ERR_FORME);
    break;
  }       
  
}

static char verifierContenuSituation(KL_Situation *kl){
  int cptPieces[KL_NB_SYMBOLES];
  int i = 0,j , ind1, ind2,ind3;
  char correct = 1;
  KL_Piece pie;
  char err[24]; /*Buffer de complement d'erreur*/
  /*Initialisation du tableau des occurences correctes de chaque symbole*/
  for (; i < (KL_NB_PIECES_1 + KL_NB_CASES_VIDES) ; i++) /*Pour les pieces de taille 1*/
    cptPieces[i] = 1; 
  for (; i < (KL_NB_PIECES_2 + KL_NB_PIECES_1 + KL_NB_CASES_VIDES); i++) /*Pour les pieces de taille 2*/
    cptPieces[i] = 2;
  cptPieces[i] = KL_TAILLE_ANE; /*Pour l'ane rouge*/
  
  /*Verification du nombre d'occurence de chaque caractere*/
  /*Et verification que les caracteres sont tous des caracteres autorises*/
  for(i = 0; i < KL_SURFACE ; i++){
    pie = getCase(*kl,i);
    if ((!estCaracPiece(pie)) && (!estCaracVide(pie))){
      sprintf(err,"'%c'",pie);
      afficherErreur(ERR_CARAC_FICHIER,err);
      correct = 0;
    }
    else{
      cptPieces[pie - 'A'] --; /*Decrementation du nombre d'occurence a trouver encore dans le fichier*/
    }
  }

  /*Affichage des erreurs detectees sur le nombre*/
  for(i = 0; i < KL_NB_SYMBOLES ; i++)
    if (cptPieces[i] != 0){ /*Si le nombre d'occurence trouvee est plus elevee que la normale, ou pas assez*/
      pie = 'A' + i;
      sprintf(err,"'%c'",pie);
      afficherErreur(ERR_NB_SYMBOLES,err);
      correct = 0;
    }  
  
  /*Verification de la forme des pieces de taille 2*/
  for(i = 0; i< KL_NB_PIECES_2  ; i++)
    if (cptPieces[i] == 0){ /*Si la piece est presente un bon nombre de fois*/
      pie = 'A' + i + KL_NB_PIECES_1 + KL_NB_CASES_VIDES ;
      j = 0;
      while(getCase(*kl,j) != pie ){ /*Recherche de la 1ere occurence de la piece*/
	j++;
      }
      /*Il suffit que, en dessous ou a droite de cet indice, la piece
       ne soit pas presente, pour que ce soit incorrect*/
      ind1= getIndDroite(j);
      ind2= getIndDessous(j);
      if ((ind1 != -1) && (getCase(*kl,ind1) != pie) && (ind2 != -1) && (getCase(*kl,ind2) != pie)){
	sprintf(err,"'%c'",pie);
	afficherErreur(ERR_FORME,err);
	correct = 0;
      }     
    }  

  /*Verification de la forme de l'ane rouge*/
  if (cptPieces[i] == 0){ /*Si le nombre d'occurences est correct*/
    j = 0;
    while(getCase(*kl,j) != KL_ANE_ROUGE ) /*Recherche de la 1ere occurence de la piece*/
      j++;
    ind1 = getIndDroite(j);
    ind2 = getIndDessous(j);
    ind3 = (ind2 == -1)? -1 : getIndDroite(ind2);
    if (((ind1 == -1) || ((ind1 != -1) && getCase(*kl,ind1)!= KL_ANE_ROUGE))
	|| ((ind2 == -1) || ((ind2 != -1) && getCase(*kl,ind2)!= KL_ANE_ROUGE))
	|| ((ind3 == -1) || ((ind3 != -1) && getCase(*kl,ind3)!= KL_ANE_ROUGE))){
      sprintf(err,"'%c'",KL_ANE_ROUGE);
      afficherErreur(ERR_FORME,err);
      correct = 0;
    }
  }
  return correct;
}

static KL_Situation *initialiserKlotski(char *lecture){
  KL_Situation *res=(KL_Situation *)allouer(sizeof(KL_Situation));
  char *s = lecture, *deblig;
  KL_Piece *k = res->surface; /*Pointeur vers le situation destination*/
  while( (s - lecture) < (KL_SURFACE+KL_HAUTEUR)){ /*Parcourt de la chaine de caracteres*/
    deblig = s;
    while ((deblig-s) < KL_LARGEUR) /*Recopie d'une ligne*/
      *k++ = *deblig++;
    s+= KL_LARGEUR + 1;
  }  
  return res;
}


KL_Situation *lireFichier(const char *nomFic){
  KL_Situation *res = NULL;
  char tampon[TAILLE_TAMPON_FIC];
  FILE *fic = fopen(nomFic,"r");
  size_t nblus;
  char correct = 1;
  if (fic == NULL) /*Si erreur d'ouverture du fichier*/
    afficherErreur(ERR_OUVERTURE,nomFic);
  else {
    tampon[TAILLE_TAMPON_FIC-1]='\0';
    nblus=fread(tampon,sizeof(char),TAILLE_TAMPON_FIC,fic);
    fclose(fic);
    if (nblus < TAILLE_CORRECTE_FIC){/*S'il n'y a pas le bon nombre de caracteres dans le fichier*/
      afficherErreur(ERR_TAILLE,nomFic);
      correct = 0;
    }
    if (correct){
      res = initialiserKlotski(tampon);
      correct = verifierContenuSituation(res) && correct;/*Affichage d'un maximum d'erreurs*/
      if (!correct){
	free(res);
	res = NULL;
      }
    }
  }
  return res;
}
