/* ***********************************************  */
/*   Projet  : Recherche de solution de l'Ane rouge */
/*   Date    : avril 2009                           */
/*   Auteur  : Climoo                               */
/*   Fichier : kl_resolution.c                      */
/*   Description : Module charge de                 */   
/*   la resolution d'une instance du jeu de         */
/*   l'Ane rouge (module central du projet).        */
/*                                                  */
/* **************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "allocations.h"
#include "fileT.h"
#include "kl_definitions.h"
#include "kl_resolution.h"
#include "kl_tabSituations.h"
#include "kl_deplacements.h"


/*Structure d'un noeud de l'arbre n_aire de recherche pour la resolution du Klotki*/
typedef struct kl_noeud{
  KL_Situation *sit;        /*Pointeur vers une situation de jeu Klotski a un instant donne*/
  struct kl_noeud *pere;    /*Pointeur vers le pere de ce noeud*/
  struct kl_noeud *filsg;   /*Pointeur vers le fils gauche de ce noeud*/
  struct kl_noeud *frered;  /*Pointeur vers le frere droit de ce noeud*/
} KL_Noeud;


/*Desalloue la memoire utilisee par un noeud*/
#define libererNoeud(noeud) (free((noeud)))

/*Defini le second noeud comme le frere (droit) du premier*/
#define ajouterFrere(n,f) ((*(n)).frered = (f))

/*Defini le second noeud comme le fils (gauche) du premier*/
#define ajouterFils(n,f) ( (*(n)).filsg = (f))

/*Defini le second noeud comme le pere du premier*/
#define ajouterPere(n,p) ((*(n)).pere = (p))

/*Fonction permettant de liberer toute la memoire utilisee par un arbre.
  Attention : les situations Klotski presentes dans les noeud ne sont pas liberees.*/
static void libererArbre(KL_Noeud *nd);

/*Fonction creant un nouveau noeud pour une situation Klotski donne.
  Les champs du fils, du premier frere droit et du pere, sont mis a NULL*/
static KL_Noeud *creerNoeud(KL_Situation *kl);

/*Fonction retournant 1 si l'ane rouge est place en position gagnante, 0 sinon*/
static char estGagnante(KL_Situation *kl);

/*Etant donne un talbeu sitfils de situation Klotski, de taille nbfils,
  la fonction permet d'ajouter au noeud nd ces situations.
  La file fnod est egalement enfilee avec chacun de ces noeuds.*/
static void ajouterSituationsNoeud(KL_Noeud *nd,File *fnod, KL_Situation **sitfils, int nbfils);

/*Fonction permettant de recopier la solution trouvee dans un tableau de mouvements,
  a partir d'un noeud, qui doit etre une feuille de l'arbre.
  Le dernier mouvement de la solution (mouvement de la feuille)sera place a l'indice 0.*/
static int recopierMouvements(KL_Noeud *feuille, KL_Mouvement **mvts);

static KL_Noeud *creerNoeud(KL_Situation *kl){
  KL_Noeud *res = (KL_Noeud*)allouer(sizeof(KL_Noeud));
  res->sit = kl;
  res->filsg = res->frered = res->pere = NULL;
  return res;
}

static void libererArbre(KL_Noeud *nd){
  if (nd->filsg != NULL)
    libererArbre(nd->filsg);
  if (nd->frered != NULL)
    libererArbre(nd->frered);
  free(nd);
  return ;
}

static char estGagnante(KL_Situation *kl){
  int ptsg[] = KL_PTS_ANE_GAGNANT ;
  int i = 0, taille = sizeof(ptsg)/ sizeof(ptsg[0]);
  while ((i < taille) && (getCase(*kl,ptsg[i])==KL_ANE_ROUGE))
    i++;
  return (i==taille);
}

static void ajouterSituationsNoeud(KL_Noeud *nd,File *fnod, KL_Situation **sitfils, int nbfils){
  KL_Noeud *nouveau, *prec;
  if (nbfils != 0){
    nouveau = creerNoeud(sitfils[--nbfils]);
    ajouterFils(nd,nouveau);
    ajouterPere(nouveau,nd);
    prec = nouveau;
    nbfils--;
    enfiler(fnod,nouveau);
    while(nbfils >= 0){
      nouveau = creerNoeud(sitfils[nbfils--]);
      ajouterFrere(prec,nouveau);
      ajouterPere(nouveau,nd);
      enfiler(fnod,nouveau);
      prec = nouveau ;
    }
  }
}

#define DEFAUT_NB_MOUVEMENTS 256

static int recopierMouvements(KL_Noeud *feuille, KL_Mouvement **mvts){
  int i = 0;
  KL_Noeud  *courant = feuille;
  *mvts = (KL_Mouvement *)allouer(DEFAUT_NB_MOUVEMENTS*sizeof(KL_Mouvement));
  while (courant->pere != NULL){
    (*mvts)[i].pie = courant->sit->mvt.pie;
    (*mvts)[i].dir = courant->sit->mvt.dir;
    courant = courant->pere;
    i++;
    if (i % DEFAUT_NB_MOUVEMENTS == 0)
      *mvts = (KL_Mouvement *)reallouer(*mvts,(i+DEFAUT_NB_MOUVEMENTS)*sizeof(KL_Mouvement));
  }
  return i;
}


void resoudre(KL_Situation *kl, KL_Solution *sol){
  KL_Situations connus; /*Tableau des situations de klotski deja connues*/
  KL_Noeud *racine = creerNoeud(kl), *courant;
  KL_Situation *sitfils[KL_NB_CASES_VIDES*KL_NB_DIRECTIONS]; 
  int nbfils;
  File fnod; /*File de noeud pour un parcourt en largeur de l'arbre*/
  sol->nbmvts = -1;
  sol->mvts = NULL;
  racine->sit = kl;
  initialiserTabSituations(&connus,kl);
  ajouterSituationConnue(connus,kl);
  initialiserFile(&fnod,0);
  enfiler(&fnod,racine);
  /*Tant qu'il reste de noeuds a visiter et tant que l'on a pas trouve une solution*/
  while( (!estVide(fnod)) && (sol->mvts == NULL) ){ 
    courant = defiler(&fnod);
    if (estGagnante(courant->sit)) /*Si on a trouve une situation gagnante*/
      sol->nbmvts = recopierMouvements(courant,&(sol->mvts));/*On recopie la solution*/
    else{
      nbfils = trouverDeplacements(connus,courant->sit,sitfils); /*Recherche des situations accessibles*/
      ajouterSituationsNoeud(courant,&fnod,sitfils,nbfils--);/*Ajout des situations accessibles*/
    }  
  }
  libererTabSituations(connus);
  supprimerFile(&fnod);
  libererArbre(racine);
}
